from pegtree.peg import *

#
# Inlining
#

def isCharOrRange(pe): 
  return isinstance(pe, PChar) or isinstance(pe, PRange)

def inline(pe: PExpr, filter=isCharOrRange):
    start = pe
    while isinstance(pe, PRef) or isinstance(pe, PName):
        pe = pe.deref()
    if filter(pe):
        if(pe != start):
            DEBUG('INLINE', start, '=>', pe)
        return pe
    return start

#
# make Minimum Rules
#

def makeMinimumRules(pe: PExpr, visited: dict, rules: list):
    if isinstance(pe, PRef):
        uname = pe.uname()
        if uname not in visited:
            visited[uname] = pe
            makeMinimumRules(pe.deref(), visited, rules)
            rules.append(pe)
        return rules
    if isinstance(pe, PUnary) or isinstance(pe, PTuple):
        for e in pe:
            makeMinimumRules(e, visited, rules)
    return rules

#
# Sorting Rules
# 

def sortRules(refs):
    newrefs = []
    unsolved = []
    for ref in refs:
        names = set([])
        _makeSortingRefs(ref.deref(), names)
        if len(names) == 0:
            newrefs.append(ref)
        else:
            unsolved.append((ref, set(names)))
    return _solveSortingRefs(newrefs, unsolved)

def _makeSortingRefs(e, names):
    if isinstance(e, PTuple):
        for e2 in e:
            _makeSortingRefs(e2, names)
    elif hasattr(e, 'e'):
        _makeSortingRefs(e.e, names)
    elif isinstance(e, PRef):
        names.add(e.uname())

def _removeSolvedName(unsolved, uname):
    removed = False
    for _, names in unsolved:
        if uname in names:
            removed = True
            names.remove(uname)
    return removed


def _solveSortingRefs(refs, unsolved):
    removed = False
    # print(refs)
    for ref in refs:
        removed |= _removeSolvedName(unsolved, ref.uname())
    max = 0
    while max < 10:
        removed = True
        while removed:
            removed = False
            newrefs = []
            stillUnsolved = []
            for ref, names in unsolved:
                if len(names) <= max:
                    refs.append(ref)
                    newrefs.append(ref)
                else:
                    stillUnsolved.append((ref, names))
            unsolved = stillUnsolved
            #print(max, newrefs)
            for ref in newrefs:
                removed |= _removeSolvedName(unsolved, ref.uname())
            if removed:
                max = 0
        max += 1
    for ref, _ in unsolved:
        refs.append(ref)
    return refs

###

def flattenSeq(pe: PExpr, ps: list, conv = lambda x: x):
  if isinstance(pe, PSeq):
    for e in pe: 
      ps.append(conv(e))
  else:
    ps.append(conv(pe))
  return ps

def appendSeq(ps: list, pe: PExpr):
  if pe == EMPTY or (isinstance(pe, PChar) and len(pe.text) == 0):
    return
  if isinstance(pe, PSeq):
    for e in pe:
      appendSeq(ps, e)
    return 
  if len(ps) == 0:
    ps.append(pe)
    return
  e0 = ps[-1]
  if isinstance(pe, PChar) and isinstance(e0, PChar):
    ps[-1] = PChar(e0.text+pe.text)
    return 
  ps.append(pe)

def newSeq(ps: list):
  optimized = []
  for e in ps:
    appendSeq(optimized, e)
  if len(optimized) == 0:
      return EMPTY
  if len(optimized) == 1:
    return optimized[0]
  return PSeq(*optimized)

#
# Ore
#

def flattenOre(pe: PExpr, ps: list, conv=lambda x: x):
  if isinstance(pe, POre):
    for e in pe:
      ps.append(conv(e))
  else:
    ps.append(conv(pe))
  return ps


def mergeRange(e, e2):
  if isAny(e) or isAny(e2):
    return ANY
  chars = ''
  ranges = ''
  if isinstance(e, PChar):
      chars += e.text
  if isinstance(e2, PChar):
      chars += e2.text
  if isinstance(e, PRange):
      chars += e.chars
      ranges += e.ranges
  if isinstance(e2, PRange):
      chars += e2.chars
      ranges += e2.ranges
  chars, ranges = uniqueRange(chars, ranges)
  return PRange.new(chars, ranges)


def prefixChar(pe):
  if isinstance(pe, PChar) and len(pe.text) > 0:
      return pe.text[0]
  if isinstance(pe, PSeq) and isinstance(pe.es[0], PChar) and len(pe.es[0].text) > 0:
      return pe.es[0].text[0]
  return None

def dc(pe):
  if isinstance(pe, PChar) and len(pe.text) > 0:
    return PChar.new(pe.text[1:])
  if isinstance(pe, PSeq) and isinstance(pe.es[0], PChar) and len(pe.es[0].text) > 0:
    first = PChar.new(pe.es[0].text[1:])
    return PSeq(first, *pe.es[1:])
  return FAIL

def appendOre(ps: list, pe, cmap=None, deref=False):
  start = pe
  while deref and (isinstance(pe, PRef) or isinstance(pe, PName)):
      pe = pe.deref()
  if isinstance(pe, POre):
      for e in pe:
          appendOre(ps, e, cmap, deref)
      return
  if len(ps) > 0:
    e0 = ps[-1]
    if isEmpty(e0): 
      return
    if isSingleCharacter(e0) and isSingleCharacter(pe):
      ps[-1] = mergeRange(e0, pe)
      return
  c = prefixChar(pe)
  if c is not None and cmap != None:
    if c not in cmap:
      cmap[c] = len(ps)
      ps.append([pe])
    else:
      nested_choice = ps[cmap[c]]
      nested_choice.append(pe)
    return
  ps.append(start)

def newOre(ps: list, cmap = None, deref = False):
    optimized = []
    for e in ps:
      appendOre(optimized, e, cmap, deref)
    for i in range(len(optimized)):
      if not isinstance(optimized[i], list):
        continue
      nested_choice = optimized[i]
      if len(nested_choice) == 1:
        optimized[i] = nested_choice[0]
      else:
        first = PChar(prefixChar(nested_choice[0]))
        nested = []
        for ne in nested_choice:
          ne = dc(ne)
          appendOre(nested, ne)
        optimized[i] = PSeq.new(first, newOre(nested, {}))
    ps = optimized
    optimized = []
    for e in ps:
      appendOre(optimized, e, None, False)
    if len(optimized) == 0:
      return FAIL
    if len(optimized) == 1:
      return optimized[0]
    if len(optimized) == 2 and isEmpty(optimized[1]):
      return POption(optimized[0])
    return POre(*optimized)

#
# Out of order execution
#

def fixedSize(e: PExpr):
    if isinstance(e, PRange) or isinstance(e, PAny):
      return 1
    if isinstance(e, PChar):
      return len(e.text)
    if isinstance(e, PAnd) or isinstance(e, PNot):
      return 0
    return -1


def splitFixed(e, conv=lambda x: x):
    ps = flattenSeq(e, [], conv)
    if fixedSize(ps[0]) == -1:
      return None, -1, e
    shift = 0
    fixed = []
    for e in ps:
      size = fixedSize(e)
      if size == -1:
          break
      shift += size
      fixed.append(e)
    #print('@', fixed, shift, ps[len(fixed):])
    return fixed, shift, ps[len(fixed):]

###

class Optimizer(object):

    def visit(self, pe: PExpr):
        pe = inline(pe)
        cname = pe.cname()
        if not hasattr(self, cname):
            return pe
        f = getattr(self, cname)
        optimized = f(pe)
        return optimized

    def PRef(self, pe):
        return pe

    def PName(self, pe):
        return PName(self.visit(pe.e), pe.name, pe.ptree, pe.isLeftRec)

    def PAnd(self, pe):
        return PAnd(self.visit(pe.e))

    def PNot(self, pe):
        return PNot(self.visit(pe.e))

    def PMany(self, pe):
        return PMany(self.visit(pe.e))

    def POneMany(self, pe):
        return POneMany(self.visit(pe.e))

    def POption(self, pe):
        return POption(self.visit(pe.e))

    def PSeq(self, pe):
        ps = flattenSeq(pe, [], lambda e: self.visit(e))
        return newSeq(ps)

    # Ore
    def POre(self, pe: POre):
        if pe.isDict() and len(pe.es) > 8: return pe
        ps = flattenOre(pe, [], lambda e: self.visit(e))
        return newOre(ps,{})

    # Tree Construction

    def PNode(self, pe: PNode):
        e = self.visit(pe.e)
        fixed, shift, ves = splitFixed(e)
        if fixed is None: # or not self.Ooox:
            return PNode(e, pe.tag, pe.shift)
        fixed.append(PNode(newSeq(ves), pe.tag, pe.shift - shift))
        return newSeq(fixed)

    def PEdge(self, pe):
        e = self.visit(pe.e)
        fixed, shift, ves = splitFixed(e)
        if fixed is None:
          return PEdge(pe.edge, e, pe.shift)
        fixed.append(PEdge(pe.edge, newSeq(ves), pe.shift - shift))
        return newSeq(fixed)

    def PFold(self, pe):
        e = self.visit(pe.e)
        fixed, shift, ves = splitFixed(e)
        if fixed is None:  # or not self.Ooox:
            return PFold(pe.edge, e, pe.tag, pe.shift)
        fixed.append(PFold(pe.edge, newSeq(ves), pe.tag, pe.shift - shift))
        return newSeq(fixed)

    def PAbs(self, pe):
        return PAbs(self.visit(pe.e))

    def PAction(self, pe):
        return PAction(self.visit(pe.e), pe.func, pe.params, pe.ptree)


optimizer = Optimizer()

def default_optimizer(e):
  return optimizer.visit(e)


# def default_optimizer(e):
#   return e

def optimize(e):
  return optimizer.visit(e)

def prepare(peg: Grammar, start_name=None, optimize_function=default_optimizer):
  #peg = peg
  if start_name is None: start_name = peg.start()
  start_ref = peg.newRef(start_name)
  refs = makeMinimumRules(start_ref, {}, [])
  refs = sortRules(refs)
  rules = {}
  memos = []
  for ref in refs:
    uname = ref.uname(peg)
    rules[uname] = optimize_function(ref.deref())
    memos.append(uname)
    # if str(rules[uname]) != str(ref.deref()):
    #   print('OPTIMIZE', ref.deref(), '\n\t=>', rules[uname])
  if 'packrat' not in peg:
    memos.clear()
  return start_ref, refs, rules, memos




from pathlib import Path
import os
from pegtree.nez import Generator
from pegtree.terminal import DefaultConsole as console

# from pegtree import Generator, grammar

PASM0 = []
PASM1 = [ 'Option', 'OneMany', 'Seq3', 'Seq4', 'Ore3', 'Ore4', 'Dict' ]
PASM1_VARGS = [ 'Seq', 'Ore' ]
PASM1_LEX = [
    'AndChar', 'NotChar', 'ManyChar', 'OneManyChar', 'OptionChar',
    'AndRange', 'NotRange', 'ManyRange', 'OneManyRange', 'OptionRange',
]

CODE = {
    'lisp': {
        'pasm': PASM1 + PASM1_LEX + PASM1_VARGS,
        'apply': '({} {})',
        'delim': ' ',
    },
    'typescript': {
        'pasm': PASM1 + PASM1_LEX + PASM1_VARGS,
    },
    'ts': {
        'file': 'ts.txt',
        'prefix': 'p',
        'rule': '  {};',
        'pasm': PASM1 + PASM1_LEX + PASM1_VARGS,
    },
}

class PAsmGenerator(Generator):
    ESCTBL = str.maketrans(
        {'\n': '\\n', '\t': '\\t', '\r': '\\r', '\v': '\\v', '\f': '\\f',
         '\\': '\\\\', "'": "\\'", '"': "\\\""})

    def __init__(self, **options):
        super().__init__()
        self.domains = []
        self.rule = '{}'
        self.apply = '{}({})'
        self.delim = ','
        self.string = '"{}"'
        self.prefix = os.environ.get('PREFIX', 'p')
        self.PASM = PASM1
        if options.get('-O', 2) == 0:
            self.PASM = PASM0
        self.rules = None

    def setup(self, spec):
        self.apply = spec.get('apply', '{}({})')
        self.string = spec.get('string', '"{}"')
        self.delim = spec.get('delim', ',')
        self.prefix = spec.get('prefix', os.environ.get('PREFIX', 'p'))
        if 'pasm' in spec and self.PASM != PASM0:
            self.PASM = set(spec['pasm'])
        self.rule = spec.get('rule', '{}')
        self.rules = []
        spec['rules'] = self.rules

    def emitRule(self, uname, pe):
        name = self.getref(uname)
        rule = self.rule.format(self.emitApply('Rule', 'peg', name, self.emit(pe, 0)))
        if isinstance(self.rules, list):
            self.rules.append(rule)
        else:
            print(rule)

    def emitParser(self, ref):
        pass

    def getref(self, name):
        if len(self.domains) > 0:
            return str(self.domains.index(name))
        return self.quote(name)

    def has(self, func):
        return func in self.PASM

    def emitApply(self, name, *args):
        name = f'{self.prefix}{name}'
        if len(args) == 0:
            return self.apply.format(name, '')
        elif len(args) == 1:
            return self.apply.format(name, args[0])
        else:
            try:
                return self.apply.format(name, self.delim.join(args))
            except:
                print('@FIXME', name, args)
                return 'FIXME'

    def quote(self, s):
        # if isinstance(s, tuple) or isinstance(s, list):
        #     sb = []
        #     for r in s:
        #         sb.append(r[0]+r[1])
        #     s = ''.join(sb)
        #     return '"' + s.translate(Parsec.ESCTBL) + '"'
        return self.string.format(str(s).translate(PAsmGenerator.ESCTBL))

    def param(self, pe):
        if hasattr(pe, 'text'):
            return (self.quote(pe.text),)
        if hasattr(pe, 'ranges'):
            return (self.quote(pe.chars), self.quote(pe.ranges))

    def PEmpty(self, pe, step):
        return self.emitApply('Empty')

    def PFail(self, pe, step):
        return self.emitApply('Fail')

    def PAny(self, pe, step):
        return self.emitApply('Any')

    def PChar(self, pe, step):
        if len(pe.text) == 0:
            return self.emitApply('Empty')
        return self.emitApply('Char', self.quote(pe.text))

    def PRange(self, pe, step):
        return self.emitApply('Range', self.quote(pe.chars), self.quote(pe.ranges))

    def PAnd(self, pe, step):
        e = pe.e
        cname = e.cname()[1:]
        if self.has(f'And{cname}'):
            return self.emitApply(f'And{cname}', *self.param(e))
        return self.emitApply('And', self.emit(e, step))

    def PNot(self, pe, step):
        e = pe.e
        cname = e.cname()[1:]
        if self.has(f'Not{cname}'):
            return self.emitApply(f'Not{cname}', *self.param(e))
        return self.emitApply('Not', self.emit(e, step))

    def PMany(self, pe, step):
        e = pe.e
        cname = e.cname()[1:]
        if self.has(f'Many{cname}'):
            return self.emitApply(f'Many{cname}', *self.param(e))
        return self.emitApply('Many', self.emit(e, step))

    def POneMany(self, pe, step):
        e = pe.e
        if self.has('OneMany'):
            cname = e.cname()[1:]
            if self.has(f'OneMany{cname}'):
                return self.emitApply(f'OneMany{cname}', *self.param(e))
            return self.emitApply('OneMany', self.emit(e, step))
        return self.emitApply('Seq2', self.emit(e, step), self.PMany(pe, pe.minLen()))

    def POption(self, pe, step):
        e = pe.e
        if self.has('Option'):
            cname = e.cname()[1:]
            if self.has(f'Option{cname}'):
                return self.emitApply(f'Option{cname}', *self.param(e))
            return self.emitApply('Option', self.emit(e, step))
        return self.emitApply('Ore2', e, self.emitApply('Empty'))

    def PSeq(self, pe, step):
        fs = []
        for e in pe:
            fs.append(self.emit(e, step))
            step += e.minLen()
        if len(fs) == 2:
            return self.emitApply('Seq2', *fs)
        if len(fs) == 3 and self.has('Seq3'):
            return self.emitApply('Seq3', *fs)
        if len(fs) == 4 and self.has('Seq4'):
            return self.emitApply('Seq4', *fs)
        if self.has('Seq'):
            return self.emitApply('Seq', *fs)
        return self.emitBin('Seq', fs)

    # Ore
    def POre(self, pe, step):
        if pe.isDict():
            ss = [s for s in pe.listDict()]
            return self.emitApply('Dict', self.quote(' '.join(ss)))
        fs = [self.emit(e, step) for e in pe]
        if len(fs) == 2:
            return self.emitApply('Ore2', *fs)
        if len(fs) == 3 and self.has('Ore3'):
            return self.emitApply('Ore3', *fs)
        if len(fs) == 4 and self.has('Ore4'):
            return self.emitApply('Ore4', *fs)
        if self.has('Ore'):
            return self.emitApply('Ore', *fs)
        return self.emitBin('Ore', fs)

    def PAlt(self, pe, step):
        return self.POre(pe, step)

    def emitBin(self, name, fs):
        if len(fs) == 1:
            return fs[0]
        if len(fs) == 2:
            return self.emitApply(f'{name}2', fs[0], fs[1])
        if self.has(f'{name}4') and len(fs) >= 4:
            return self.emitApply(f'{name}4', fs[0], fs[1], fs[2], self.emitBin(name, fs[3:]))
        if self.has(f'{name}3') and len(fs) >= 3:
            return self.emitApply(f'{name}3', fs[0], fs[1], self.emitBin(name, fs[2:]))
        return self.emitApply(f'{name}2', fs[0], self.emitBin(name, fs[1:]))

    def PRef(self, pe, step):
        ref = self.getref(pe.uname(self.peg))
        return self.emitApply('Ref', 'peg', ref)

    # Tree Construction

    def PNode(self, pe, step):
        e = self.emit(pe.e, step)
        return self.emitApply('Node', e, self.quote(pe.tag), f'{pe.shift}')

    def PEdge(self, pe, step):
        e = self.emit(pe.e, step)
        return self.emitApply('Edge', self.quote(pe.edge), e, f'{pe.shift}')

    def PFold(self, pe, step):
        e = self.emit(pe.e, step)
        return self.emitApply('Fold', self.quote(pe.edge), e, self.quote(pe.tag), f'{pe.shift}')

    def PAbs(self, pe, step):
        e = self.emit(pe.e, step)
        return self.emitApply('Abs', e)

    def Skip(self, pe, step):
        return self.emitApply('Skip')

    def Symbol(self, pe, step):
        sid = self.getsid(str(pe.params[0]))
        e = self.emit(pe.e, step)
        return self.emitApply('Symbol', e, str(sid))

    def Exists(self, pe, step):
        sid = self.getsid(str(pe.params[0]))
        return self.emitApply('Exists', str(sid))

    def Match(self, pe, step):
        sid = self.getsid(str(pe.params[0]))
        return self.emitApply('Match', str(sid))

    def Scope(self, pe, step):
        e = self.emit(pe.e, step)
        return self.emitApply('Scope', e)

    def Def(self, pe, step):
        params = pe.params
        name = str(params[1]) if len(params) == 2 else str(params[0])
        e = self.emit(pe.e, step)
        return self.emitApply('Def', self.quote(name), e)

    def In(self, pe, step):
        params = pe.params
        name = str(params[0])
        return self.emitApply('In', self.quote(name))


def output(spec):
    if 'file' in spec:
        path = Path(__file__).resolve().parent / 'code' / spec['file']
        with path.open() as f:
            for line in f:
                if line.startswith('#@RULE'):
                    for rule in spec['rules']:
                        print(rule)
                else:
                    print(line, end='')
    else:
        for rule in spec['rules']:
            print(rule)


def parsec(peg, **options):
    generator = PAsmGenerator(**options)
    if 'format' in options:
        format = options['format']
        if format not in CODE:
            names = sorted(CODE.keys())
            print(console.bold('Supported format:'), ' '.join(names))
            return
        spec = CODE[format]
        generator.setup(spec)
        generator.generate(peg, **options)
        output(spec)
    else:
        generator.generate(peg, **options)




from collections import namedtuple


def pRule(peg, name, pf):
    peg[name] = pf

# Generator

def pEmpty():
    return match_empty

def pFail():
    return match_fail

def pAny():
    return match_any

def match_empty(px): 
    return True

def match_fail(px): 
    return False

def match_any(px):
    if px.pos < px.epos:
        px.pos += 1
        return True
    return False

## Characters

Name2ParseFunc = {
    '': (match_empty, match_empty)
}

Range2ParseFunc = {
}

def find_match(pfunc):
    for func, funcmove in Name2ParseFunc.values():
        if pfunc == funcmove:
            return func
    for func, funcmove in Range2ParseFunc.values():
        if pfunc == funcmove:
            return func
    return None

def string_pfunc(text):
    clen = len(text)
    def match_char(px):
        if px.inputs.startswith(text, px.pos):
            px.pos += clen
            return True
        return False
    match_andchar = lambda px: px.inputs.startswith(text, px.pos)
    Name2ParseFunc[text] = (match, match_move)
    return Name2ParseFunc[text]

## range

def range_keys_append(keys, s, e):
    if s > e:
        s, e = e, s
    for i in range(len(keys)):
        skey, ekey = keys[i]
        if skey <= s and s <= ekey:
            if ekey < e:
                keys[i] = (skey, e)
            return
        elif skey <= e and e <= ekey:
            if s < skey:
                keys[i] = (s, ekey)
            return
        elif ord(ekey)+1 == ord(s):
            keys[i] = (skey, e)
        elif ord(e)+1 == ord(skey):
            keys[i] = (s, ekey)
    keys.append((s, e))

def range_unique(chars, ranges):
    keys=[]
    for c in chars:
        range_keys_append(keys,c,c)
    for i in range(0, len(ranges), 2):
        range_keys_append(keys,ranges[i],ranges[i+1])
    chars = []
    ranges = []
    for s, e in sorted(keys):
        if s == e:
            chars.append(s)
        else:
            ranges.append(s+e)
    return ''.join(chars), ''.join(ranges)

def move1(pf):
    def move(px):
        if pf(px):
            px.pos+=1
            return True
        return False
    return move

def range_pfunc(chars, ranges):
    chars, ranges = range_unique(chars, ranges)
    key = (chars, ranges)
    if key in Range2ParseFunc:
        return Range2ParseFunc[key]
    long_ranges = []
    for i in range(0, len(ranges), 2):
        s, e = ord(ranges[i]), ord(ranges[i+1])
        if e - s < 255:
            chars += ''.join(ord(c) for c in range(s, e+1))
        else:
            long_ranges.append(ranges[i])
            long_ranges.append(ranges[i+1])
    r = ''.join(long_ranges)
    if len(r) == 0:
        def match(px):
            return px.pos < px.epos and chars.find(px.inputs[px.pos]) != -1
        Range2ParseFunc[key] = (match, move1(match))
    elif len(long_ranges) == 2:
        def match(px):
            if px.pos < px.epos:
                c = px.inputs[px.pos]
                return chars.find(c) != -1 or r[0] <= c <= r[1]
            return False
        Range2ParseFunc[key] = (match, move1(match))
    elif len(long_ranges) == 4:
        def match(px):
            if px.pos < px.epos:
                c = px.inputs[px.pos]
                return chars.find(c) != -1 or r[0] <= c <= r[1] or r[2] <= c <= r[3]
            return False
        Range2ParseFunc[key] = (match, move1(match))
    else:
        def match(px):
            if px.pos < px.epos:
                c = px.inputs[px.pos]
                if chars.find(c) != -1: return True
                for i in range(0, r, 2):
                    if r[i] <= c <= r[i+1]: return True
            return False
        Range2ParseFunc[key] = (match, move1(match))
    return Range2ParseFunc[key]

def pChar(text):
    _, pf = string_pfunc(text)
    return pf

def pRange(chars, ranges):
    _, pf = range_pfunc(chars, ranges)
    return pf

## and

def pfunc(string_or_pfunc):
    if isinstance(string_or_pfunc, str):
        return pChar(string_or_pfunc)
    return string_or_pfunc

def pAnd(pf):
    pf = pfunc(pf)
    text = find_string(pf)
    if text is not None:
        return pAndChar(text)
    pfa = find_andmatch(pf)
    if match is not None:
        return match
    def match_and(px):
        pos = px.pos
        if pf(px):
            px.headpos = max(px.pos, px.headpos)
            px.pos = pos
            return True
        return False
    return match_and

def pAndChar(text):
    def match_andchar(px):
        return px.inputs.startswith(text, px.pos)
    return match_andchar

def pNot(pf):
    text = get_string(pf)
    if text is not None:
        return pNotChar(text)
    def match_not(px):
        pos = px.pos
        ptree = px.ptree
        if not pf(px):
            px.headpos = max(px.pos, px.headpos)
            px.pos = pos
            px.ptree = ptree
            return True
        return False
    return match_not

def pNotChar(text):
    def match_notchar(px):
        return not px.inputs.startswith(text, px.pos)
    return match_notchar

def pMany(pf):
    text = get_string(pf)
    if text is not None:
        return pManyChar(text)
    def match_many(px):
        pos = px.pos
        ptree = px.ptree
        while pf(px) and pos < px.pos:
            pos = px.pos
            ptree = px.ptree
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ptree = ptree
        return True
    return match_many

def pManyChar(text):
    clen = len(text)

    def match_manychar(px):
        while px.inputs.startswith(text, px.pos):
            px.pos += clen
        return True
    return match_manychar

def pOneMany(pf):
    text = get_string(pf)
    if text is not None:
        return pOneManyChar(text)

    def match_OneMany(px):
        if pf(px):
            pos = px.pos
            ptree = px.ptree
            while pf(px) and pos < px.pos:
                pos = px.pos
                ptree = px.ptree
            px.headpos = max(px.pos, px.headpos)
            px.pos = pos
            px.ptree = ptree
            return True
        return False
    return match_OneMany

def pOneManyChar(text):
    clen = len(text)
    def match_OneManychar(px):
        if px.inputs.startswith(text, px.pos):
            px.pos += clen
            while px.inputs.startswith(text, px.pos):
                px.pos += clen
            return True
        return False
    return match_OneManychar

def pOption(pf):
    text = get_string(pf)
    if text is not None:
        return pOptionChar(text)
    def match_option(px):
        pos = px.pos
        ptree = px.ptree
        if not pf(px):
            px.headpos = max(px.pos, px.headpos)
            px.pos = pos
            px.ptree = ptree
        return True
    return match_option

def pOptionChar(text):
    clen = len(text)
    def match_optionchar(px):
        if px.inputs.startswith(text, px.pos):
            px.pos += clen
        return True
    return match_optionchar

# Seq

def pSeq(*pfs):
    if len(pfs) == 1:
        return pfs[0]
    elif len(pfs) == 2:
        return pSeq2(pfs[0], pfs[1])
    elif len(pfs) == 3:
        return pSeq3(pfs[0], pfs[1], pfs[2])
    elif len(pfs) == 4:
        return pSeq4(pfs[0], pfs[1], pfs[2], pfs[3])
    def match_seq(px):
        for pf in pfs:
            if not pf(px):
                return False
        return True
    return match_seq


def pSeq2(pf, pf2):
    def match_seq2(px):
        return pf(px) and pf2(px)
    return match_seq2


def pSeq3(pf, pf2, pf3):
    def match_seq3(px):
        return pf(px) and pf2(px) and pf3(px)
    return match_seq3


def pSeq4(pf, pf2, pf3, pf4):
    def match_seq4(px):
        return pf(px) and pf2(px) and pf3(px) and pf4(px)
    return match_seq4



# Ore

def pOre(*pfs):
    if len(pfs) == 1:
        return pfs[0]
    elif len(pfs) == 2:
        return pOre2(pfs[0], pfs[1])
    elif len(pfs) == 3:
        return pOre3(pfs[0], pfs[1], pfs[2])
    elif len(pfs) == 4:
        return pOre4(pfs[0], pfs[1], pfs[2], pfs[3])

    def match_ore(px):
        pos = px.pos
        ptree = px.ptree
        for pf in pfs:
            if pf(px):
                return True
            px.headpos = max(px.pos, px.headpos)
            px.pos = pos
            px.ptree = ptree
        return False
    return match_ore

def pOre2(pf, pf2):
    def match_ore2(px):
        pos = px.pos
        ptree = px.ptree
        if pf(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ptree = ptree
        return pf2(px)
    return match_ore2


def pOre3(pf, pf2, pf3):
    def match_ore3(px):
        pos = px.pos
        ptree = px.ptree
        if pf(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ptree = ptree
        if pf2(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ptree = ptree
        return pf3(px)
    return match_ore3


def pOre4(pf, pf2, pf3, pf4):
    def match_ore4(px):
        pos = px.pos
        ptree = px.ptree
        if pf(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ptree = ptree
        if pf2(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ptree = ptree
        if pf3(px):
            return True
        px.headpos = max(px.pos, px.headpos)
        px.pos = pos
        px.ptree = ptree
        return pf4(px)
    return match_ore4




# def make_trie(dic):
#     if '' in dic or len(dic) < 10:
#         return dic
#     d = {}
#     for s in dic:
#         s0, s = s[0], s[1:]
#         if s0 in d:
#             ss = d[s0]
#             if not s in ss:
#                 ss.append(s)
#         else:
#             d[s0] = [s]
#     for key in d:
#         d[key] = make_trie(d[key])
#     return d


# def match_trie(px, d):
#     if px.pos >= px.epos:
#         return False
#     if isinstance(d, dict):
#         c = px.inputs[px.pos]
#         if c in d:
#             px.pos += 1
#             return match_trie(px, d[c])
#         return False
#     pos = px.pos
#     inputs = px.inputs
#     for s in d:
#         if inputs.startswith(s, pos):
#             px.pos += len(s)
#             return True
#     return False


# def pDict(words):
#     if isinstance(words, str):
#         words = words.split(' ')
#     dic = make_trie(words)
#     return lambda px: match_trie(px, dic)


# def pRef(generated, uname):
#     if uname not in generated:
#         fs = None

#         def match_deref(px):
#             nonlocal fs
#             if fs is None:
#                 fs = generated[uname]
#             return fs(px)
#         generated[uname] = match_deref
#     return generated[uname]


def pRef(generated, uname):
    if uname not in generated:
        generated[uname] = lambda px: generated[uname](px)
    return generated[uname]

# def pLeft(pf):
#     def match(px):
#         key = (px.pos, uname)
#         if key in px.rec:
#             return False
#         px.rec[key] = True
#         result = pf
#         del px.rec[key]
#         return result
#     return match

# def pDyRef(uname, default_pf=match_fail):
#     def match(px):
#         pf = px.pfmap.get(uname, default_pf)
#         return pf(px)
#     return match

# def pApply(pf, name, param):
#     def match(px):
#         px.pfmap[name] = param
#         return pf(px)
#     return match


class PMemo(object):
    __slots__ = ['key', 'pos', 'treeState', 'ptree', 'prev', 'result']

    def __init__(self):
        self.key = -1
        self.pos = 0
        self.treeState = False
        self.ptree = None
        self.prev = None
        self.result = False


def pMemo(fs, mp, mpsize):
    disabled = False
    hit = 0
    miss = 0

    def match_memo(px):
        nonlocal disabled, hit, miss
        if disabled:
            return fs(px)
        key = (mpsize * px.pos) + mp
        m = px.memo[key % 1789]
        if m.key == key:
            if m.treeState:
                if m.prev == px.ptree:
                    px.pos = m.pos
                    px.ptree = m.ptree
                    hit += 1
                    return m.result
            else:
                px.pos = m.pos
                hit += 1
                return m.result
        prev = px.ptree
        m.result = fs(px)
        m.pos = px.pos
        m.key = key
        if m.result and prev != px.ptree:
            m.treeState = True
            m.prev = prev
            m.ptree = px.ptree
        else:
            m.treeState = False
        miss += 1
        if miss % 100 == 0:
            if hit / miss < 5:
                disabled = True
        return m.result
    return match_memo


def pMemoDebug(name, fs, mp, mps):
    disabled = False
    hit = 0
    miss = 0
    mpsize = len(mps)

    def match_memo(px):
        nonlocal disabled, hit, miss
        if disabled:
            return fs(px)
        key = (mpsize * px.pos) + mp
        m = px.memo[key % 1789]
        if m.key == key:
            if m.treeState:
                if m.prev == px.ptree:
                    px.pos = m.pos
                    px.ptree = m.ptree
                    hit += 1
                    return m.result
            else:
                px.pos = m.pos
                hit += 1
                return m.result
        prev = px.ptree
        m.result = fs(px)
        m.pos = px.pos
        m.key = key
        if m.result and prev != px.ptree:
            m.treeState = True
            m.prev = prev
            m.ptree = px.ptree
        else:
            m.treeState = False
        miss += 1
        if miss % 100 == 0:
            if hit / miss < 5:
                mps.remove(name)
                print('enabled', mps)
                disabled = True
        return m.result
    return match_memo


# Tree Construction


class PTree(object):
    __slots__ = ['prev', 'tag', 'spos', 'epos', 'child']

    def __init__(self, prev, tag, spos, epos, child):
        self.prev = prev
        self.tag = tag
        self.spos = spos
        self.epos = epos
        self.child = child

    def isEdge(self):
        return self.epos < 0

    def dump(self, inputs):
        sb = []
        if self.prev is not None:
            sb.append(self.prev.dump(inputs))
            sb.append(',')
        sb.append(f'{{#{self.tag} ')
        if self.child is None:
            sb.append(repr(inputs[self.spos:self.epos]))
        else:
            sb.append(self.child.dump(inputs))
        sb.append('}')
        return ''.join(sb)

# Node (tag, spos, epos, child)
# Edge (prev, edge, node)

def pNode(pf, tag, shift):
    def make_node(px):
        pos = px.pos
        prev = px.ptree
        px.ptree = None
        if pf(px):
            node = (tag, pos+shift, px.pos, px.ptree)
            px.ptree = (prev, '', node)
            return True
        return False
    return make_node

def pEdge(edge, pf, shift=0):
    def make_edge(px):
        pos = px.pos
        prev = px.ptree
        if pf(px):
            if px.ptree is prev:
                node = (tag, pos+shift, px.pos, None)
                px.ptree = (prev, edge, node)
            else:
                p = px.tree
                px.ptree = (p[1], edge, p[2])
            return True
        return False
    return make_edge

def make_ptree(urn, inputs, tag, spos, epos, ptree):
    if ptree is None:
        return ParseTree(tag, inputs, spos, epos, urn)
    prev, edge, node = ptree
    if prev is None and tag == '':
        return make_ptree(urn, inputs, *node)
    t = ParseTree(tag, inputs, spos, epos, urn)
    t.append(edge, make_ptree(urn, inputs, *node))
    while prev is not None:
        prev, edge, node = ptree
        t.append(edge, make_ptree(urn, inputs, *node))
    for i in range(len(t)//2):
        t[i], t[-(1+i)] = t[-(1+i)], t[i]
    return t

# def popPTree(px):
#     pt = px.ptree
#     if pt is None:
#         return px.pos, None
#     if pt.prev is None:
#         px.ptree = pt
#         return pt.spos, None
#     px.ptree = PTree(None, pt.tag, pt.spos, pt.epos, pt.child)
#     return pt.spos, pt.prev

def pFold(edge, pf, tag, shift):
    def match_fold(px):
        if px.ptree is None:
            spos = px.pos
            prev = None
            px.ptree = None
        else:
            prev, _, ptree = px.ptree
            spos = ptree.spos
            px.ptree = (None, edge, ptree)
        if pf(px):
            node = (tag, spos, px.pos, px.ptree)
            px.ptree = (prev, '', node)
            return True
        return False
    return match_fold

def pAbs(pf):
    def match_abs(px):
        ptree = px.ptree
        if pf(px):
            px.ptree = ptree
            return True
        return False
    return match_abs

def rechain(mprev, mtree, ptree):
    prev, tree = mtree
    if prev is mprev:
        prev 

def pMemo(fs, mp, mpsize):
    disabled = False
    hit = 0
    miss = 0

    def match_tree(px):
        nonlocal disabled, hit, miss
        if disabled:
            return fs(px)
        key = (mpsize * px.pos) + mp
        m = px.memo[key % 1789]
        if m.key == key:
            hit += 1
            if m.treeModified:
                if m.prev is px.ptree:
                    px.pos = m.pos
                    px.ptree = m.ptree
                    return m.result
                else:
                    px.tree = rechain(m.prev, px.tree, m.tree)
            else:
                px.pos = m.pos
                hit += 1
                return m.result
        prev = px.ptree
        m.result = fs(px)
        m.pos = px.pos
        m.key = key
        if m.result and prev is not px.ptree:
            m.treeModified = True
            m.prev = prev
            m.ptree = px.ptree
        else:
            m.treeModified = False
        miss += 1
        if miss % 100 == 0:
            if hit / miss < 5:
                disabled = True
        return m.result
    return match_memo



def pSkip():  # @skip()
    def skip(px):
        px.pos = min(px.headpos, px.epos)
        return True
    return skip

# State


State = namedtuple('State', 'sid val prev')


def getstate(state, sid):
    while state is not None:
        if state.sid == sid:
            return state
        state = state.prev
    return None


def pSymbol(pf, sid):  # @symbol(A)
    def match_symbol(px):
        pos = px.pos
        if pf(px):
            px.state = State(sid, px.inputs[pos:px.pos], px.state)
            return True
        return False
    return match_symbol


def pScope(pf):
    def scope(px):
        state = px.state
        res = pf(px)
        px.state = state
        return res
    return scope


def pExists(sid):  # @Match(A)
    return lambda px: getstate(px.state, sid) != None


def pMatch(sid):  # @Match(A)
    def match(px):
        state = getstate(px.state, sid)
        if state is not None and px.inputs.startswith(state.val, px.pos):
            px.pos += len(state.val)
            return True
        return False
    return match

# params = pe.params
# name = str(params[0])
# pf = self.emit(pe.e, step)


def pDef(name, pf):
    def define_dic(px):
        pos = px.pos
        if pf(px):
            s = px.inputs[pos:px.pos]
            if len(s) == 0:
                return True
            if name not in px.dic:
                px.dic[name] = []
            ss = px.dic[name]
            ss.append(s)
            px.dic[name] = sorted(ss, key=lambda x: len(x))[::-1]
            # print(px.dic[name])
            return True
        return False
    return define_dic


# params = pe.params
# name = str(params[0])

def pIn(name):  # @in(NAME)
    def match_dic(px):
        # print('@matching', name, px.inputs, px.pos)
        if name in px.dic:
            ss = px.dic[name]
            for s in ss:
                if px.inputs.startswith(s, px.pos):
                    # print('@matched', s)
                    px.pos += len(s)
                    return True
            # print('@', px.inputs[px.pos:], ss)
        return False
    return match_dic

def pNotRange(chars, ranges):
    pf = pAndRange(chars, ranges)

    def match_ranges(px):
        return not pf(px)
    return match_ranges


def pManyRange(chars, ranges):
    pf = pAndRange(chars, ranges)

    def match_manybitset(px):
        while pf(px):
            px.pos += 1
        return True
    return match_manybitset


def pOneManyRange(chars, ranges):
    pf = pAndRange(chars, ranges)

    def match_onemanybitset(px):
        c = 0
        while pf(px):
            px.pos += 1
            c += 1
        return c > 0
    return match_onemanybitset


def pOptionRange(chars, ranges):
    pf = pAndRange(chars, ranges)

    def match_optionbitset(px):
        if pf(px):
            px.pos += 1
        return True
    return match_optionbitset

# generate


# PContext


class PContext:
    __slots__ = ['inputs', 'pos', 'epos',
                 'headpos', 'ptree', 'state', 'memo', 'dic']

    def __init__(self, inputs, spos, epos):
        self.inputs = inputs
        self.pos = spos
        self.epos = epos
        self.headpos = spos
        self.ptree = None
        self.state = None
        self.memo = [PMemo() for x in range(1789)]
        self.dic = {}

# ParseTree


def rowcol(urn, inputs, spos):
    inputs = inputs[:spos + (1 if len(inputs) > spos else 0)]
    rows = inputs.split(b'\n' if isinstance(inputs, bytes) else '\n')
    return urn, spos, len(rows), len(rows[-1])-1


def nop(s): return s


UNKNOWN_SOURCE = '(unknown source)'




def PTreeConv(pt: PTree, urn, inputs):
    if pt.prev != None:
        ct = pt
        while ct.prev is not None:
            ct = ct.prev
        return PTreeConvNode('', urn, inputs, ct.spos, pt.epos, pt)
    if pt.isEdge():
        ct = PTreeConv(pt.child, urn, inputs)
        t = ParseTree('', inputs, ct.spos_, ct.epos_, urn)
        t.set(pt.tag, ct)
        return t
    else:
        return PTreeConvNode(pt.tag, urn, inputs, pt.spos, pt.epos, pt.child)


def PTreeConvNode(tag, urn, inputs, spos, epos, subnode):
    t = ParseTree(tag, inputs, spos, epos, urn)
    while subnode != None:
        if subnode.isEdge():
            # if subnode.child == None:
            #     tt = PTreeConvNode(
            #         '', urn, inputs, subnode.spos, abs(subnode.epos), None)
            # else:
            assert subnode.child != None
            tt = PTreeConv(subnode.child, urn, inputs)
            t.set(subnode.tag, tt)
        else:
            t.append(PTreeConvNode(subnode.tag, urn, inputs,
                                   subnode.spos, abs(subnode.epos), subnode.child))
        subnode = subnode.prev
    for i in range(len(t)//2):
        t[i], t[-(1+i)] = t[-(1+i)], t[i]
    return t

def treeConvAsList(urn, inputs, ptree):
    if len(ptree) > 2:
        tag, spos, epos, child = ptree

    if len(ptree) == 2: #(prev, node)
        t = ParseTree(tag, inputs, spos, epos, urn)
        prev, node = ptree
        tag, spos, epos, 


def generate(pf):
    # pf = self.generated[start.uname()]
    def parse(inputs, urn='(unknown source)', pos=0, epos=None, conv=PTreeConv):
        if epos is None:
            epos = len(inputs)
        px = PContext(inputs, pos, epos)
        if not pf(px):
            result = PTree(None, "err", px.headpos, px.headpos, None)
        else:
            result = px.ptree if px.ptree is not None else PTree(None,
                                                                 "", pos, px.pos, None)
        return conv(result, urn, inputs)
    return parse



from setuptools import setup
import sys
from pathlib import Path
sys.path.append(str(Path(__file__).resolve().parent / 'tests'))

'''
python3 -m unittest
vim setup.py
rm -rf dist/
python3 setup.py sdist bdist_wheel
twine upload --repository pypi dist/*
'''

setup(
    name='pegtree',
    version='0.9.14',
    url='https://github.com/KuramitsuLab/pegtree.git',
    license='MIT',
    author='Kimio Kuramitsu',
    description='PEGTree for Python',
    install_requires=['setuptools'],
        packages=['pegtree'],
        package_data={'pegtree': ['grammar/*.tpeg', 'grammar/*.pegtree',
                                  'grammar/*/*.txt',
                                  'code/*.txt', 'cjdic/*.txt']},
    entry_points={
        'console_scripts': [
            'pegtree = pegtree.main:main'
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing',
    ],
    test_suite='test_all.suite'
)



import unittest
from pegtree.peg import *


class TestPeg(unittest.TestCase):

    def test_Empty(self):
        pe = PChar('')
        # print(pe)
        self.assertFalse(isAlwaysConsumed(pe))

    def test_Char(self):
        pe = PChar('a')
        # print(pe)
        self.assertTrue(isAlwaysConsumed(pe))

    def test_String(self):
        pe = PChar('abc')
        # print(pe)
        self.assertTrue(isAlwaysConsumed(pe))

    def test_Range(self):
        pe = PRange('0123456789', 'AZaz')
        # print(pe)
        self.assertTrue(isAlwaysConsumed(pe))

    def test_Any(self):
        pe = PAny()
        # print(pe)
        self.assertTrue(isAlwaysConsumed(pe))

    def test_Seq(self):
        pe = PSeq(PChar('abc'), PRange('', 'AZaz'))
        # print(pe)
        self.assertTrue(isAlwaysConsumed(pe))

    def test_And(self):
        pe = PAnd(PChar('a'))
        # print(pe)
        self.assertFalse(isAlwaysConsumed(pe))

    def test_Not(self):
        pe = PNot(PChar('a'))
        # print(pe)
        self.assertFalse(isAlwaysConsumed(pe))

    def test_Many(self):
        pe = PMany(PChar('a'))
        # print(pe)
        self.assertFalse(isAlwaysConsumed(pe))

    def test_OneMany(self):
        pe = POneMany(PChar('a'))
        # print(pe)
        self.assertTrue(isAlwaysConsumed(pe))

    def test_Option(self):
        pe = POption(PChar('a'))
        # print(pe)
        self.assertFalse(isAlwaysConsumed(pe))

    def test_Ref(self):
        peg = Grammar()
        peg['A'] = PChar('a')
        pe = peg.newRef('A')
        # print(pe)
        self.assertTrue(isAlwaysConsumed(pe))

    # def test_Choice(self):
    #     print(PE.mergeRange(PChar('A'),
    #                         PRange('0', 'AZ'),
    #                         ))
    #     print(PE.mergeRange(PChar('a'),
    #                         PChar('b'),
    #                         ))
    #     print(PE.mergeRange(PChar('a'),
    #                         PChar('c'),
    #                         ))
    #     es = [
    #         PChar('return'),
    #         PChar('r'),
    #         PChar('ab'),
    #         PChar('a'),
    #         PChar('b'),
    #         PChar('ba'),
    #         PChar('ret'),
    #     ]
    #     print(PE.newChoice(es))
    #     es = [
    #         PChar('r'),
    #         PChar('a'),
    #         PChar('c'),
    #         PChar('b'),
    #     ]
    #     print(PE.newChoice(es))

    # def test_Char(self):
    #     parser = generate(pChar('a'))
    #     tree = parser('aa')
    #     self.assertEqual(str(tree), 'a')
    #     self.assertEqual(repr(tree), "[# 'a']")

    # def test_Node(self):
    #     parser = generate(pNode(pChar('a'), 'A', 0))
    #     tree = parser('aa')
    #     self.assertEqual(str(tree), 'a')
    #     self.assertEqual(repr(tree), "[#A 'a']")

    # def test_NestedNode(self):
    #     parser = generate(pNode(pSeq2(pNode(pChar('a'), 'A', 0),
    #                                   pNode(pChar('b'), 'B', 0)), "AB", 0))
    #     tree = parser('ab')
    #     self.assertEqual(str(tree), 'ab')
    #     self.assertEqual(repr(tree), "[#AB [#A 'a'][#B 'b']]")

    # def test_FoldNode(self):
    #     parser = generate(pSeq2(pNode(pChar('a'), 'A', 0),
    #                             pFold('', pChar('b'), 'B', 0)))
    #     tree = parser('ab')
    #     self.assertEqual(str(tree), 'ab')
    #     self.assertEqual(repr(tree), "[#B [#A 'a']]")

    # def test_NodeNode(self):
    #     parser = generate(pSeq2(pNode(pChar('a'), 'A', 0),
    #                             pNode(pChar('b'), 'B', 0)))
    #     tree = parser('ab')
    #     self.assertEqual(str(tree), 'ab')
    #     self.assertEqual(repr(tree), "[# [#A 'a'][#B 'b']]")
    #     print(tree.dump())


if __name__ == '__main__':
    unittest.main()



from pegtree.peg import *
import pegtree.pasm as pasm
from pegtree.optimizer import prepare
from pegtree.terminal import DefaultConsole as console

class Generator(object):
    def __init__(self):
        self.peg = None
        self.generated = {}
        self.generating_nonterminal = ''
        self.sids = {}
        self.memos = []
        self.Ooox = True
        self.Olex = True

    def getsid(self, name):
        if not name in self.sids:
            self.sids[name] = len(self.sids)
        return self.sids[name]

    def generate(self, peg, **option):
        self.peg = peg
        self.generated = {}
        start, refs, rules, memos = prepare(peg, option.get('start', peg.start()))
        self.memos = memos
        for ref in refs:
            assert isinstance(ref, PRef)
            uname = ref.uname(peg)
            self.generating_nonterminal = uname
            self.emitRule(uname, rules[uname])
            self.generating_nonterminal = ''
        return self.emitParser(start)

    def emitRule(self, uname, pe: PExpr):
        A = self.emit(pe, 0)
        if uname in self.memos:
            idx = self.memos.index(uname)
            if idx != -1:
                A = pasm.pMemo(A, idx, len(self.memos))
                # A = pasm.pMemoDebug(ref.name, A, idx, self.memos)
        self.generated[uname] = A

    def emitParser(self, start):
        return pasm.generate(self.generated[start.uname(self.peg)])

    def emit(self, pe: PExpr, step: int):
        cname = pe.cname()
        if hasattr(self, cname):
            f = getattr(self, cname)
            return f(pe, step)
        print('@TODO(Generator)', cname, pe)
        return self.PChar(EMPTY, step)

    def PAny(self, pe, step):
        return pasm.pAny()

    def PChar(self, pe, step):
        return pasm.pChar(pe.text)

    def PRange(self, pe, step):
        return pasm.pRange(pe.chars, pe.ranges)

    def PAnd(self, pe, step):
        e = pe.e
        if(self.Olex and isinstance(e, PChar)):
            return pasm.pAndChar(e.text)
        if(self.Olex and isinstance(e, PRange)):
            return pasm.pAndRange(e.chars, e.ranges)
        return pasm.pAnd(self.emit(e, step))

    def PNot(self, pe, step):
        e = pe.e
        if(self.Olex and isinstance(e, PChar)):
            return pasm.pNotChar(e.text)
        if(self.Olex and isinstance(e, PRange)):
            return pasm.pNotRange(e.chars, e.ranges)
        return pasm.pNot(self.emit(e, step))

    def PMany(self, pe, step):
        e = pe.e
        if(self.Olex and isinstance(e, PChar)):
            return pasm.pManyChar(e.text)
        if(self.Olex and isinstance(e, PRange)):
            return pasm.pManyRange(e.chars, e.ranges)
        return pasm.pMany(self.emit(e, step))

    def POneMany(self, pe, step):
        e = pe.e
        if(self.Olex and isinstance(e, PChar)):
            return pasm.pOneManyChar(e.text)
        if(self.Olex and isinstance(e, PRange)):
            return pasm.pOneManyRange(e.chars, e.ranges)
        return pasm.pOneMany(self.emit(e, step))

    def POption(self, pe, step):
        e = pe.e
        if(self.Olex and isinstance(e, PChar)):
            return pasm.pOptionChar(e.text)
        if(self.Olex and isinstance(e, PRange)):
            return pasm.pOptionRange(e.chars, e.ranges)
        return pasm.pOption(self.emit(e, step))

    def PSeq(self, pe, step):
        pfs = []
        for e in pe:
            pfs.append(self.emit(e, step))
            step += e.minLen()
        pfs = tuple(pfs)
        if len(pfs) == 2:
            return pasm.pSeq2(pfs[0], pfs[1])
        if len(pe) == 3:
            return pasm.pSeq3(pfs[0], pfs[1], pfs[2])
        if len(pe) == 4:
            return pasm.pSeq4(pfs[0], pfs[1], pfs[2], pfs[3])
        return pasm.pSeq(*pfs)

    # Ore
    def POre(self, pe: POre, step):
        if pe.isDict():
            return pasm.pDict(pe.listDict())
        pfs = tuple(map(lambda e: self.emit(e, step), pe))
        if len(pfs) == 2:
            return pasm.pOre2(pfs[0], pfs[1])
        if len(pe) == 3:
            return pasm.pOre3(pfs[0], pfs[1], pfs[2])
        if len(pe) == 4:
            return pasm.pOre4(pfs[0], pfs[1], pfs[2], pfs[3])
        return pasm.pOre(*pfs)

    def PRef(self, pe, step):
        return pasm.pRef(self.generated, pe.uname(self.peg))

    def PName(self, pe: PName, step):
        if step == 0 and self.generating_nonterminal == pe.e.uname(self.peg):
            console.perror(pe.position, 'Left recursion')
            return self.emit(FAIL, step)
        return self.PRef(pe.e, step)

    # Tree Construction

    def PNode(self, pe, step):
        fs = self.emit(pe.e, step)
        return pasm.pNode(fs, pe.tag, pe.shift)

    def PEdge(self, pe, step):
        return pasm.pEdge(pe.edge, self.emit(pe.e, step), pe.shift)

    def PFold(self, pe, step):
        fs = self.emit(pe.e, step)
        return pasm.pFold(pe.edge, fs, pe.tag, pe.shift)

    def PAbs(self, pe, step):
        return pasm.pAbs(self.emit(pe.e, step))

    def Skip(self, pe, step):  # @skip()
        return pasm.pSkip()

    def Symbol(self, pe, step):  # @symbol(A)
        params = pe.params
        sid = self.getsid(str(params[0]))
        return pasm.pSymbol(self.emit(pe.e, step), sid)

    def Scope(self, pe, step):
        return pasm.pScope(self.emit(pe.e, step))

    def Exists(self, pe, step):  # @Match(A)
        params = pe.params
        sid = self.getsid(str(params[0]))
        return pasm.pExists(sid)

    def Match(self, pe, step):  # @Match(A)
        params = pe.params
        sid = self.getsid(str(params[0]))
        return pasm.pMatch(sid)

    def Def(self, pe, step):  # @def(A, '名詞')
        params = pe.params
        name = str(params[1]) if len(params) == 2 else str(params[0])
        pf = self.emit(pe.e, step)
        #print('@def', pf, name)
        return pasm.pDef(name, pf)

    def In(self, pe, step):  # @in(A)
        params = pe.params
        name = str(params[0])
        #print('@in', name)
        return pasm.pIn(name)


generator = Generator()


def generate(peg, **options):
    return generator.generate(peg, **options)



import unittest
from pegtree.optimizer import *


class TestOptimizer(unittest.TestCase):

    def test_Node(self):
        pe = PNode(PChar('a'), 'A', 0)
        po = optimize(pe)
        #print(pe, '=>', po)
        self.assertEqual(str(po), "'a' {/*-1*/ '' #A }")

    def test_Node2(self):
        pe = PNode(PChar('abc'), 'A', 0)
        po = optimize(pe)
        #print(pe, '=>', po)
        self.assertEqual(str(po), "'abc' {/*-3*/ '' #A }")

    def test_NodeMany(self):
        pe = PNode(PMany(PChar('abc')), 'A', 0)
        po = optimize(pe)
        #print(pe, '=>', po)
        self.assertEqual(str(po), "{ 'abc'* #A }")

    def test_NodeMany(self):
        pe = PNode(PSeq(PChar('a'), PMany(PChar('b'))), 'A', 0)
        po = optimize(pe)
        #print(pe, '=>', po)
        self.assertEqual(str(po), "'a' {/*-1*/ 'b'* #A }")

    def test_NodeNot(self):
        pe = PNode(PNot(PChar('a')), 'A', 0)
        po = optimize(pe)
        #print(pe, '=>', po)
        self.assertEqual(str(po), "!'a' { '' #A }")

    def test_NodeNode(self):
        pe = PNode(PNode(PChar('a'), 'A', 0), 'B', 0)
        po = optimize(pe)
        #print(pe, '=>', po)
        self.assertEqual(str(po), "'a' {/*-1*/ {/*-1*/ '' #A } #B }")

    def test_Edge(self):
        pe = PEdge('x', PChar('a'), 0)
        po = optimize(pe)
        #print(pe, '=>', po)
        self.assertEqual(str(po), "'a' x/*-1*/: ''")

    def test_Fold(self):
        pe = PFold('', PChar('a'), 'A', 0)
        po = optimize(pe)
        #print(pe, '=>', po)
        self.assertEqual(str(po), "'a' { ^/*-1*/ '' #A }")

    def test_RangeOrRange(self):
        pe = POre(PRange('abc', ''), PRange('ABC', ''))
        po = optimize(pe)
        #print(pe, '=>', po)
        self.assertEqual(str(po), "[A-Ca-c]")

    def test_RangeOrChar(self):
        pe = POre(PRange('abc', ''), PChar('A'))
        po = optimize(pe)
        #print(pe, '=>', po)
        self.assertEqual(str(po), "[Aa-c]")

    def test_RangeOrAny(self):
        pe = POre(PRange('abc', ''), ANY)
        po = optimize(pe)
        #print(pe, '=>', po)
        self.assertEqual(str(po), ".")

    def test_Dict(self):
        pe = POre(PChar('int'), PChar('if'))
        po = optimize(pe)
        #print(pe, '=>', po)
        self.assertEqual(str(po), "'i' ('nt' / 'f')")

    def test_Dict2(self):
        pe = POre(PChar('int'), PChar('i'), PChar('if'))
        po = optimize(pe)
        #print(pe, '=>', po)
        self.assertEqual(str(po), "'i' 'nt'?")

    def test_Dict3(self):
        pe = POre(PChar('int'), PChar('a'), PChar('b'), PChar('if'))
        po = optimize(pe)
        #print(pe, '=>', po)
        self.assertEqual(str(po), "'i' ('nt' / 'f') / [ab]")


if __name__ == '__main__':
    unittest.main()



from .pasm import ParseTree

from logging import getLogger
logger = getLogger(__name__)

class Visitor(object):
    def __init__(self, methods=None):
        self.methods = methods if methods is not None else {}

    def visit(self, tree: ParseTree):
        tag = tree.getTag()
        methods = self.methods
        if tag not in methods:
            methods[tag] = f'accept{tag}'
        method = methods[tag]
        if hasattr(self, method):
            method = getattr(self, method)
            return method(tree)
        else:
            logger.warning(
                f'@TODO undefined {method} method for {repr(tree)}')
            return self.acceptUndefined(tree)

    def acceptUndefined(self, tree):
        return None

    def TODO(self, msg, tree):
        logger.warning(f'@TODO {msg} for {repr(tree)}')

    def perror(self, ptree, msg='syntax error'):
        logger.error(ptree.message(msg))

    def warning(self, ptree, msg):
        logger.warning(ptree.message(msg))



import os
from pegtree.pasm import unique_range, ParseTree

## 
from logging import getLogger
logger = getLogger(__name__)

def perror(sp: ParseTree, msg: str):
    if sp is not None:
        logger.warning(sp.message(msg))

# Parsing Expression

class PExpr(object):

    def __iter__(self): 
        pass

    def __len__(self): 
        return 0
    
    def isNullable(self):
        return True
    
    def hasParseTree(self): 
        return False
    
    def match(self, x):
        return None

# NonTerminal

def isNullableName(name):
    return name[-1] == '_'


def isCamelStyleName(name):
    upper = 0
    lower = 0
    for c in list(name):
        if c.isupper():
            upper += 1
        elif c.islower():
            lower += 1
    return upper > 0 and lower > 0


class PRef(PExpr):
    name: str
    peg: object # Grammar
    tree: ParseTree
    isLeftRec: bool
    isLazy: bool

    def __init__(self, peg, name, tree=None):
        self.peg = peg
        self.name = name
        self.tree = None
        self.isLeftRec = False
        self.isLazy = False

    def __repr__(self):
        return self.name

    def uname(self, peg=None):
        if self.peg == peg:
            return self.name
        return f'{self.peg.ns}{self.name}'

    def deref(self):
        if self.name not in self.peg:
            store_default_parsing_expression(self.peg, self.name, self.tree)           
        return self.peg[self.name]

    def isNullable(self):
        return isNullableName(self.name)

    def hasParseTree(self): 
        return not isCamelStyleName(self.name)

    def match(self, x):
        return self.deref().match(x)


DefaultNonTerminals = {
}

def store_default_parsing_expression(peg, name, sp: ParseTree):
    if len(DefaultNonTerminals) == 0:
        DefaultNonTerminals = {
            '_': PMany(PRange(' \t\u3000', '')),
            '__': PMany(PRange(' \t\r\n\u3000', '')),
            'DIGIT': PRange('', '09'),
            'ALPHA': PRange('', 'AZaz'),
        }
    if name.startswith('"') and name.startswith('"'):
        string_literal = PSeq.new(PChar.new(name[1:-1]), PRef(peg, '_'))
        peg[name] = string_literal
    elif ord(name[0]) > 127: # utf文字
        peg[name] = PChar.new(name)
    elif name in DefaultNonTerminals:
        peg[name] = DefaultNonTerminals[name]
    else:
        if sp is not None:
            perror(sp, f'undefined name: {name}')
        peg[name] = EMPTY if isNullableName(name) else FAIL 

# Charater, Literal 

class PAny(PExpr):

    def __repr__(self): 
        return '.'

    def isNullable(self):
        return False

    def match(self, x):
        return x[1:] if len(x) > 0 else None


class PChar(PExpr):
    text: str 
    ESCTBL = str.maketrans(
        {'\n': '\\n', '\t': '\\t', '\r': '\\r', '\v': '\\v', '\f': '\\f',
         '\\': '\\\\', "'": "\\'"})

    def __init__(self, text):
        self.text = text

    def __repr__(self):
        return "'" + self.text.translate(PChar.ESCTBL) + "'"

    def isNullable(self):
        return len(self.text) == 0

    def match(self, x):
        return x[len(self.text):] if x.startswith(self.text) else None

    @classmethod
    def new(cls, text):
        if len(text) == 0:
            return EMPTY
        return PChar(text)


class PRange(PExpr):
    __slots__ = ['chars', 'ranges', 'neg']
    ESCTBL = str.maketrans(
        {'\n': '\\n', '\t': '\\t', '\r': '\\r', '\v': '\\v', '\f': '\\f',
         '\\': '\\\\', ']': '\\]', '-': '\\-'})

    def __init__(self, chars, ranges, neg=False):
        self.chars = chars
        self.ranges = ranges
        self.neg = neg

    def __repr__(self):
        sb = []
        if(self.neg):
            sb.append('(!')
        sb.append('[')
        sb.append(self.chars.translate(PRange.ESCTBL))
        r = self.ranges
        while len(r) > 1:
            sb.append(r[0].translate(PRange.ESCTBL))
            sb.append('-')
            sb.append(r[1].translate(PRange.ESCTBL))
            r = r[2:]
        sb.append(']')
        if(self.neg):
            sb.append('. )')
        return ''.join(sb)

    def isNullable(self):
        return False

    def match(self, x):
        if len(x) > 0:
            c = x[0]
            if c in self.chars: return True
            for i in range(0, len(self.ranges), 2):
                if self.ranges[i] <= c <= self.ranges[i+1]:
                    return True
        return False

    @classmethod
    def new(cls, chars, ranges, neg=False):
        if len(chars) == 0 and len(ranges) == 0:
            return ANY if neg else FAIL
        if len(chars) == 1 and len(ranges) == 0 and neg == False:
            return PChar(chars)
        return PRange(chars, ranges, neg)


## Unary

def repr_unary(e):
    if isinstance(e, POre) or isinstance(e, PSeq) or isinstance(e, PEdge) or isinstance(e, PAlt):
        return '(' + repr(e) + ')'
    return repr(e)


class PUnary(PExpr):
    e: PExpr

    def __init__(self, e):
        self.e = e

    def __iter__(self):
        yield self.e

    def __len__(self):
        return 1

    def isNullable(self): 
        return self.e.isNullable()

    def hasParseTree(self): 
        return self.e.hasParseTree()

    def match(self, x):
        return self.e.match(x)


# Parsing Expression

class PNot(PUnary):
    e: PExpr

    def __repr__(self):
        return '!'+repr_unary(self.e)

    def isNullable(self): 
        return True

    def isParseTree(self): 
        return False

    def match(self, x):
        return x if self.e.match(x) is None else None


class PAnd(PUnary):
    e: PExpr

    def __repr__(self):
        return '&'+repr_unary(self.e)

    def isNullable(self): 
        return True

    def match(self, x):
        return None self.e.match(x) is None else x


class PMany(PUnary):
    e: PExpr

    def __repr__(self):
        return repr_unary(self.e)+'*'

    def isNullable(self): 
        return True

    def match(self, x):
        y = self.e.match(x)
        while y is not None and len(y) < len(x):
            x = y
            y = self.e.match(x)
        return x


class POneMany(PUnary):
    e: PExpr

    def __repr__(self):
        return repr_unary(self.e)+'+'

    def isNullable(self): 
        return self.e.isNullable()

    def match(self, x):
        y = self.e.match(x)
        if y is not None:
            return PMany.match(self, y)
        return None


class POption(PUnary):
    e: PExpr

    def __repr__(self):
        return repr_unary(self.e)+'?'

    def match(self, x):
        y = self.e.match(x)
        return x if y is None else y

## Binary

def repr_seq(e):
    if isinstance(e, POre) or isinstance(e, PAlt):
        return '(' + repr(e) + ')'
    return repr(e)


class PTuple(PExpr):
    es: tuple

    def __init__(self, *es):
        self.es = tuple(es)

    def __iter__(self):
        return iter(self.es)

    def __len__(self):
        return len(self.es)


class PSeq(PTuple):
    def __repr__(self):
        return ' '.join(map(lambda e: repr_seq(e), self.es))

    def isNullable(self):
        for e in self.es:
            if not e.isNullable(): return False
        return True

    def hasParseTree(self):
        for e in self.es:
            if e.hasParseTree: return True
        return False

    def match(self, x):
        for e in self.es:
            x = e.match(x)
            if x is None: return None
        return x

    @classmethod
    def new(cls, *es):
        if len(es) == 0:
            return EMPTY
        return es[0] if len(es) == 1 else PSeq(*es)


class POre(PTuple):
    def __repr__(self):
        return ' / '.join(map(repr, self))
    
    def isNullable(self):
        for e in self.es:
            if e.isNullable(): return True
        return False

    def hasParseTree(self):
        for e in self.es:
            if e.hasParseTree(): return True
        return False

    def match(self, x):
        for e in self.es:
            y = e.match(x)
            if x is not None: 
                return y
        return None

    def isDict(self):
        for e in self:
            if not isinstance(e, PChar):
                return False
        return True

    def listDict(self):
        dic = [e.text for e in self if isinstance(e, PChar)]
        dic2 = []
        for s in dic:
            if s == '':
                break
            dic2.append(s)
        return dic2

    @classmethod
    def new(cls, *es):
        if len(es) == 0:
            return FAIL
        return es[0] if len(es) == 1 else POre(*es)

class PAlt(POre):
    def __repr__(self):
        return ' | '.join(map(repr, self))

# Tree Construction 

def repr_tree_tag(tag):
    return ' ' if tag == '' else f' #{tag} '

def repr_tree_edge(edge, suffix=''):
    return suffix if edge == '' else f'{edge}:{suffix}'

def repr_tree_shift(shift):
    return '' if shift == 0 else f'/*{shift}*/'


class PNode(PUnary):
    e: PExpr
    tag: str
    shift: int

    def __init__(self, e, tag='', shift=0):
        self.e = e
        self.tag = tag
        self.shift = shift

    def __repr__(self):
        shift = repr_tree_tag(self.shift)
        tag = repr_tree_shift(self.tag)
        return '{' + shift + ' ' + repr(self.e) + tag + '}'

    def hasParseTree(self):
        return True
    


class PEdge(PUnary):
    edge: str
    e: PExpr
    shift: int

    def __init__(self, edge, e, shift=0):
        self.e = e
        self.edge = edge
        self.shift = shift

    def __repr__(self):
        shift = repr_tree_tag(self.shift)
        edge = repr_tree_edge(self.edge)
        return shift + edge + repr_unary(self.e)

    def hasParseTree(self):
        return True


class PFold(PUnary):
    edge: str
    e: PExpr
    tag: str
    shift: int

    def __init__(self, edge, e, tag='', shift=0):
        self.e = e
        self.edge = edge
        self.tag = tag
        self.shift = shift

    def __repr__(self):
        shift = repr_tree_shift(self.shift)
        edge = repr_tree_edge(self.edge, suffix='^ ')
        tag = repr_tree_tag(self.tag)
        return '{'+ shift + edge + repr(self.e) + tag + '}'

    def hasParseTree(self):
        return True


class PAbs(PUnary):
    e: PExpr

    def __init__(self, e):
        self.e = e

    def __repr__(self):
        return f'@abs({self.e})'

    def hasParseTree(self):
        return False

# Symbol

class PSymbol(PUnary):
    e: PExpr

    def __init__(self, e):
        self.e = e

    def __repr__(self):
        return f'@symbol({self.e})'

class PMatch(PUnary):
    e: PExpr

    def __init__(self, e):
        self.e = e

    def __repr__(self):
        return f'@match({self.e})'

class PAny(PUnary):
    e: PExpr

    def __init__(self, e):
        self.e = e

    def __repr__(self):
        return f'@any({self.e})'

class PRepeat(PUnary):
    e: PExpr
    name: str

    def __init__(self, e):
        self.e = e

    def __repr__(self):
        return f'@repeat({self.e}, {self.name})'

class PScope(PUnary):
    e: PExpr

    def __init__(self, e):
        self.e = e

    def __repr__(self):
        return f'@scope({self.e})'



# Action


class PAction(PUnary):
    __slots__ = ['e', 'func', 'params', 'ptree']

    def __init__(self, e, func, params, ptree=None):
        self.e = e
        self.func = func
        self.params = params
        self.ptree = ptree

    def __repr__(self):
        return f'@{self.func}{self.params}'

    def cname(self):
        return self.func.capitalize()

# CONSTANT
EMPTY = PChar('')
ANY = PAny()
FAIL = PNot(EMPTY)


def isEmpty(pe):
  return pe == EMPTY or isinstance(pe, PChar) and len(pe.text) == 0


def isAny(pe):
  return pe == ANY or isinstance(pe, PAny)


def isSingleCharacter(pe):
        return (isinstance(pe, PChar) and len(pe.text) == 1) or isinstance(pe, PRange) or isinstance(pe, PAny)


class Visitor(object):
    visited: dict
    def __init__(self, visited):
        self.visited = visited

    def visit(self, pe: PExpr):
        if self.visited is not None and isinstance(pe, PExpr):
            if pe.name in self.visited:
                return self.visited[pe.name]
        key = f'accept{pe.__class__.__name__}'
        if hasattr(self, key):
            f = getattr(self, key)
            return f(pe)
        if isinstance(pe, PExpr):
            return self.visitPUnary(pe)
        if isinstance(pe, PTuple):
            return self.visitPTuple(pe)
        return self.visitUndefined(pe)

    def visitPUnary(self, pe):
        return self.visitUndefined(pe)

    def visitPTuple(self, pe):
        return self.visitUndefined(pe)

    def visitUndefined(self, pe):
        return self.visitUndefined(pe)

        




    

# PRange Utilities

def bitsetRange(chars, ranges):
    cs = 0
    for c in chars:
        cs |= 1 << ord(c)
    r = ranges
    while len(r) > 1:
        for c in range(ord(r[0]), ord(r[1])+1):
            cs |= 1 << c
        r = r[2:]
    return cs


def stringfyRange(bits):
    c = 0
    s = None
    p = None
    chars = []
    ranges = []
    while bits > 0:
        if bits & 1 == 1:
            if s is None:
                s = c
                p = c
            elif p + 1 == c:
                p = c
            else:
                _appendRange(s, p, chars, ranges)
                s = c
                p = c
        bits >>= 1
        c += 1
    if s is not None:
        _appendRange(s, p, chars, ranges)
    return ''.join(chars), ''.join(ranges)


def _appendRange(s, p, chars, ranges):
    if s == p:
        chars.append(chr(s))
    elif s+1 == p:
        chars.append(chr(s))
        chars.append(chr(p))
    else:
        ranges.append(chr(s))
        ranges.append(chr(p))

def uniqueRange(chars, ranges):
    bits = bitsetRange(chars, ranges)
    newchars, newranges = stringfyRange(bits)
    checkbits = bitsetRange(chars, ranges)
    assert bits == checkbits
    return newchars, newranges


def checkLeftRec(peg, name, e=None, visited=None):
    if e == None:
        visited = set()
        e = peg[name]
    if isinstance(e, PRef):
        if name == e.name:
            e.isLeftRec = True
            return
        if e.name not in visited:
            visited.add(e.name)
            checkLeftRec(peg, name, e.deref(), visited)
    elif isinstance(e, PSeq):
        for ei in e.es:
            checkLeftRec(peg, name, ei, visited)
            if not ei.isNullable(): return
    elif len(e) != 0:
        for ei in e:
            checkLeftRec(peg, name, ei, visited)


    

#
# Visitor
# 

class PVisitor(object):
    def visit(self, pe: PExpr):
        key = self.cname(pe)
        f = getattr(self, key)
        return f(pe)

    def cname(self, pe):
        return pe.cname()


class Nullable(PVisitor):
    def PChar(self, pe): return 1 if len(pe.text) > 0 else 0
    def PAny(self, pe): return 1
    def PRange(self, pe): return 1

    def PRef(self, pe):
        if not hasattr(pe, 'nullable') or pe.nullable is None:
            pe.nullable = 0
            pe.nullable = self.visit(pe.deref())
        return pe.nullable

    def PAnd(self, pe): return 0
    def PNot(self, pe): return 0
    def PMany(self, pe): return 0
    def POneMany(self, pe): return self.visit(pe.e)
    def POption(self, pe): return 0

    def PSeq(self, pe):
        for e in pe:
            if(self.visit(e) == 1):
                return 1
        return 0

    def POre(self, pe):
        for e in pe:
            if(self.visit(e) == 0):
                return 0
        return 1

    def PNode(self, pe): return self.visit(pe.e)
    def PFold(self, pe): return self.visit(pe.e)
    def PEdge(self, pe): return self.visit(pe.e)
    def PAbs(self, pe): return self.visit(pe.e)

    def PAction(self, pe): return self.visit(pe.e)

defaultNullableChecker = Nullable()

def isAlwaysConsumed(pe):
    return defaultNullableChecker.visit(pe)

class First(PVisitor):

    def __init__(self):
        self.memos = {}

    def PChar(self, pe):
        if len(pe.text) == 0:
            return (0, 0)
        return (1 << ord(pe.text[0]), 0)

    def PAny(self, pe):
        return (-1, 0)

    def PRange(self, pe):
        if(pe.neg):
            return (-1, unique_range(pe.chars, pe.ranges))
        return (unique_range(pe.chars, pe.ranges), 0)

    def PRef(self, pe):
        uname = pe.uname()
        if(uname not in self.memos):
            self.memos[uname] = (-1, 0)
            self.memos[uname] = self.visit(pe.deref())
        return self.memos[uname]

    def PNot(self, pe):
        if isSingleCharacter(pe.e):
            pos, neg = self.visit(pe.e)
            return (neg, pos)
        return LeftRef.EMPTYSET

    def PAnd(self, pe): return self.visit(pe.e)
    def PMany(self, pe): return self.visit(pe.e)
    def POneMany(self, pe): return self.visit(pe.e)
    def POption(self, pe): return self.visit(pe.e)

    def PSeq(self, pe):
        pos, neg = 0, 0
        for e in pe:
            pos1, neg1 = self.visit(e)
            pos |= pos1
            neg |= neg1
            if isAlwaysConsumed(e):
                break
        return (pos, neg)

    def POre(self, pe):
        pos, neg = 0, 0
        for e in pe:
            pos1, neg1 = self.visit(e)
            pos |= pos1
            neg |= neg1
        return (pos, neg)

    def PNode(self, pe): return self.visit(pe.e)
    def PFold(self, pe): return self.visit(pe.e)
    def PEdge(self, pe): return self.visit(pe.e)
    def PAbs(self, pe): return self.visit(pe.e)

    def PAction(self, pe): return self.visit(pe.e)


# Grammar

GrammarId = 0


class Grammar(dict):
    def __init__(self, source=None):
        global GrammarId
        self.ns = str(GrammarId)
        self.N = []
        GrammarId += 1
        super().__setitem__('@@example', [])
        super().__setitem__('@@refs', {})

    def __repr__(self):
        ss = []
        for rule in self.N:
            ss.append(rule)
            ss.append(' = ')
            ss.append(repr(self[rule]))
            ss.append('\n')
        return ''.join(ss)

    def __setitem__(self, key, item):
        if not key in self:
            self.N.append(key)
        super().__setitem__(key, item)

    def newRef(self, name):
        refs = self['@@refs']
        if name not in refs:
            refs[name] = PRef(self, name)
        return refs[name]

    def start(self):
        if len(self.N) == 0:
            self['EMPTY'] = EMPTY
        return self.N[0]


